import http.server
from pathlib import Path


PORT = 20776

Handler = lambda *args: http.server.SimpleHTTPRequestHandler(
    *args, directory=str(Path(__file__).parent / "webapp")
)


def serve(org, token, host="localhost", port=PORT):
    with http.server.ThreadingHTTPServer(("", PORT), Handler) as httpd:
        print(
            f"Telemetry app URL: http://{host}:{port}/index.html?org={org}&token={token}&host={host}"
        )

        httpd.serve_forever()



import typing as t
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from enum import Enum

from f1.packets import TRACKS
from f1.packets import CarStatusData
from f1.packets import LapData
from f1.packets import PacketSessionData
from f1.packets import SessionType


class SessionState(Enum):
    INIT = 0
    IN_GARAGE = 1
    ON_TRACK = 2
    FINISHED = 3


class SessionEventHandler(ABC):
    @abstractmethod
    def on_new_session(self, session: "Session") -> None: ...

    @abstractmethod
    def on_new_lap(
        self,
        current_lap: int,
        previous_lap: int,
        previous_sectors: t.Tuple[int, int, int],
        best: bool,
    ) -> None: ...

    @abstractmethod
    def on_sector(self, n: int, lap: int, time: float, best: bool) -> None: ...

    @abstractmethod
    def on_finish(
        self, lap: int, sectors: t.Tuple[int, int, int], best: bool
    ) -> None: ...


class Session:
    def __init__(self, handler: SessionEventHandler) -> None:
        self.handler = handler

        self.state = SessionState.INIT

        self.lap = 0
        self.sector = 0
        self.sectors = [None, 0, 0, 0]
        self.best_lap_time = 0
        self.best_sectors = [None, 0, 0, 0]
        self.best_lap_sectors = [None, 0, 0, 0]
        self.tyre: t.Optional[str] = None
        self.tyre_age = None
        self.slug: t.Optional[str] = None
        self.session_uid = None
        self.track = None
        self.type = None
        self.time = None  # seconds
        self.fuel = None

        self._lap_data = None

    def step(self):
        self.state = getattr(self, f"handle_{self.state.name}")()

    def handle_INIT(self):
        if self._lap_data is None:
            return SessionState.INIT

        return (
            SessionState.ON_TRACK
            if self._lap_data.driver_status in (1, 4)
            else SessionState.IN_GARAGE
        )

    def _update_sector_3(self):
        if self._lap_data is None:
            return

        total_time = self._lap_data.last_lap_time_in_ms
        sector_time = self.sectors[3] = (
            0
            if any(_ == 0 for _ in self.sectors[1:3])
            else total_time - sum(self.sectors[1:3])
        )
        best = self.best_sectors[3] == 0 or self.best_sectors[3] > sector_time
        if best:
            self.best_sectors[3] = sector_time

        return best

    def _update_last_lap(self) -> bool:
        if self._lap_data is None:
            return False

        total_time = self._lap_data.last_lap_time_in_ms
        best = self.best_lap_time == 0 or self.best_lap_time > total_time
        if best:
            self.best_lap_time = total_time
            self.best_lap_sectors = tuple(self.sectors)
        return best

    def handle_IN_GARAGE(self):
        if self._lap_data is None:
            return SessionState.INIT

        self.lap = self._lap_data.current_lap_num
        self.sector = self._lap_data.sector + 1

        return (
            SessionState.ON_TRACK
            if self._lap_data.driver_status != 0
            else SessionState.IN_GARAGE
        )

    def handle_ON_TRACK(self):
        if self._lap_data is None:
            return SessionState.INIT

        try:
            current_lap = self._lap_data.current_lap_num
            current_sector = self._lap_data.sector + 1

            if self._lap_data.driver_status == 0:
                return SessionState.IN_GARAGE

            if (self.lap, self.sector) < (current_lap, current_sector):
                # Flashback
                pass

            if self.lap < current_lap:  # new lap
                best = self._update_sector_3()

                self.handler.on_sector(3, self.lap, self.sectors[3], best)

                best = self._update_last_lap()

                self.handler.on_new_lap(
                    current_lap, self.lap, tuple(self.sectors[1:]), best
                )

            elif self.sector < current_sector:  # new sector
                sector_time = getattr(
                    self._lap_data, f"sector{self.sector}_time_ms_part"
                )
                sector_time += (
                    getattr(self._lap_data, f"sector{self.sector}_time_minutes_part")
                    * 60
                    * 1000
                )
                self.sectors[self.sector] = sector_time
                best = (
                    self.best_sectors[self.sector] == 0
                    or self.best_sectors[self.sector] > sector_time
                )
                if best:
                    self.best_sectors[self.sector] = sector_time

                self.handler.on_sector(self.sector, self.lap, sector_time, best)

            return SessionState.ON_TRACK

        finally:
            self.lap = current_lap
            self.sector = current_sector

    def handle_FINISHED(self):
        best = self._update_sector_3()
        self.handler.on_sector(3, self.lap, self.sectors[3], best)

        best = self._update_last_lap()
        self.handler.on_finish(self.lap, tuple(self.sectors[1:]), best)

        self.session_uid = None

        return SessionState.INIT

    def refresh(self, packet: PacketSessionData):
        """Refresh the current session."""
        self.time = packet.header.session_time

        if self.session_uid == packet.header.session_uid:
            return

        self.__init__(self.handler)

        self.session_uid = packet.header.session_uid
        self.track = TRACKS[packet.track_id]
        self.slug = f'{datetime.now().strftime("%Y-%m-%d|%H:%M")}|{self.track}'
        self.type = packet.session_type

        self.handler.on_new_session(self)

        self.step()

    def lap_data(self, data: LapData):
        self._lap_data = data

        self.step()

    def car_status_data(self, data: CarStatusData):
        self.tyre = {16: "Soft", 17: "Medium", 18: "Hard", 7: "Inter", 8: "Wet"}.get(
            data.visual_tyre_compound, "Unknown"
        )
        self.tyre_age = data.tyres_age_laps
        self.fuel = data.fuel_remaining_laps

    def final_classification(self):
        self.state = SessionState.FINISHED

        self.step()

    def is_qualifying(self):
        if self.type is None:
            return False

        return SessionType.Q <= self.type <= SessionType.OSSQ

    def is_race(self):
        if self.type is None:
            return False

        return SessionType.RACE <= self.type <= SessionType.RACE_3



import asyncio
import json

import websockets


LIVE_QUEUE = None
loop = asyncio.get_event_loop()
_CONNECTIONS = 0


def enqueue(data):
    global LIVE_QUEUE

    if LIVE_QUEUE is None:
        return False

    loop.call_soon_threadsafe(LIVE_QUEUE.put_nowait, data)

    return True


async def consume_queue(websocket):
    global LIVE_QUEUE, _CONNECTIONS

    _CONNECTIONS += 1

    if LIVE_QUEUE is None:
        LIVE_QUEUE = asyncio.Queue()

    try:
        while True:
            try:
                data = await asyncio.wait_for(LIVE_QUEUE.get(), 0.5)
            except asyncio.TimeoutError:
                await websocket.ping()
                continue

            await websocket.send(json.dumps(data))

    except websockets.exceptions.ConnectionClosed:
        _CONNECTIONS -= 1

        if _CONNECTIONS == 0:
            LIVE_QUEUE = None
            print("Queue cleared")
        return


async def _serve(host="localhost", port=20775):
    async with websockets.serve(consume_queue, host, port):
        await asyncio.Future()  # run forever


def serve(host="localhost", port=20775):
    loop.run_until_complete(_serve(host, port))



import threading
import typing as t
from bisect import bisect_left
from collections import defaultdict
from collections import deque
from time import time

import pyttsx3
from f1.handler import PacketHandler
from f1.packets import TYRES
from f1.packets import Packet
from f1.packets import PacketCarDamageData
from f1.packets import PacketCarTelemetryData
from f1.packets import PacketEventData
from f1.packets import PacketFinalClassificationData
from f1.packets import PacketLapData
from f1.packets import PacketParticipantsData
from f1.packets import PacketSessionData
from f1.packets import SessionType

from f1_telemetry.live import enqueue
from f1_telemetry.model import Session
from f1_telemetry.model import SessionEventHandler
from f1_telemetry.report import HumanCounter
from f1_telemetry.report import QualifyingReport
from f1_telemetry.report import RaceDirector
from f1_telemetry.report import RaceReport
from f1_telemetry.view import SessionPrinter


def closest(values, value):
    pos = bisect_left(values, value)
    if pos == 0:
        return pos
    if pos == len(values):
        return pos - 1
    before = values[pos - 1]
    after = values[pos]
    if after - value < value - before:
        return pos
    return pos - 1


def _flatten_tyre_values(data, name):
    temps = data.pop(name)
    data.update(
        {f"{name}_{tyre.name.lower()}": temp for tyre, temp in zip(TYRES, temps)}
    )


def _player_index(packet: Packet) -> int:
    return packet.header.player_car_index


def player_name(player) -> str:
    name = player.name.decode()
    return f"{name}{player.network_id}" if name == "Player" else name


def _weather(packet: PacketSessionData) -> t.Tuple[str, str]:
    return {
        0: ("☀️", "Clear"),
        1: ("🌥️", "Light cloud"),
        2: ("☁️", "Overcast"),
        3: ("🌦️", "Light rain"),
        4: ("🌧️", "Heavy rain"),
        5: ("⛈️", "Storm"),
    }.get(packet.weather, ("?", packet.weather))


class TelemetryCollector(PacketHandler, SessionEventHandler):
    def __init__(self, listener, sink, report=False):
        super().__init__(listener)

        self.sink = sink
        self.queue = deque()  # To handle flashbacks

        self.session = Session(self)
        self.motion_data = None
        self.tyre_data_emitted = False

        self.last_live_data = {}
        self.printer = SessionPrinter()

        self.gap = 0.0  # meters
        self.leader_distance = []
        self.leader_time = []
        self.leader_timestamp = []
        self.rival_index = 255
        self.distance = 0.0
        self.rival_distance = 0.0

        self.report = report
        self.drivers = {} if report else None
        self.human_count = HumanCounter() if report else None

        self.max_tyre_wear = None
        self.stop_needed = False
        self.wing_status = (0, 0)

        self._last_forecast = None

        self.last_packets = {}

        self.sc_slow_down_timestamp = 0

        self.rival_motion_data = None

        self.race_director = None

        self.last_fuel = None

    def push(self, fields: t.Dict[str, t.Any]):
        current_time = self.session.time
        if (
            self.session is None
            or self.session.lap == 0
            or self.distance < 0
            or current_time is None
        ):
            return

        self.queue.append((current_time, self.session.lap, fields))

        # Flush data outside the flashback window
        flashback_time = current_time - 16.0
        while self.queue:
            time, lap, data = self.queue[0]
            if time >= flashback_time:
                break
            self.queue.popleft()
            self.sink.write(f"{self.session.slug}|{lap:002}", data)

    def flush(self):
        while self.queue:
            _, lap, data = self.queue.popleft()
            self.sink.write(f"{self.session.slug}|{lap:002}", data)

    def push_live(self, _type: str, data: t.Dict[str, t.Any]):
        if self.session is None:
            return

        live_data = {"type": _type, "data": data}
        if live_data == self.last_live_data.get(_type, None):
            return

        if enqueue({"type": _type, "data": data}):
            self.last_live_data[_type] = live_data

    def collect(self):
        return self.handle()

    # ---- SessionEventHandler ----

    def on_sector(self, n: int, lap: int, time: float, best: bool) -> None:
        self.push({f"sector_{n}_ms": time})
        if time <= 0:
            return

        best_time = sum(self.session.best_lap_sectors[1 : n + 1])
        current_time = sum(self.session.sectors[1 : n + 1])

        self.printer.print_sector(
            n, lap, time, best, best_time == 0 or current_time < best_time
        )

    def on_new_lap(self, current_lap, previous_lap, previous_sectors, best):
        damage_packet = self.get_last(PacketCarDamageData)

        last_lap_time = sum(previous_sectors)
        rate_per_lap = 0

        # Estimate tyre wear rate
        current_max_wear = max(
            damage_packet.car_damage_data[_player_index(damage_packet)].tyres_wear
        )
        if self.max_tyre_wear is not None:
            rate_per_lap = current_max_wear - self.max_tyre_wear
            if rate_per_lap > 0:
                remaining_laps = (
                    self.get_last(PacketSessionData).total_laps - previous_lap
                )
                laps_to_pit = int((75 - current_max_wear) / rate_per_lap)
                self.stop_needed = laps_to_pit < remaining_laps
                if 0 < laps_to_pit < min(5, remaining_laps):
                    self.engineer(
                        f"Pit in {laps_to_pit} laps. {laps_to_pit} laps till pit"
                        if laps_to_pit > 1
                        else "BOX BOX BOX!"
                    )
                elif current_max_wear > 25:
                    self.engineer(
                        f"Tyre wear {round(current_max_wear)}. Degradation rate {round(rate_per_lap)}."
                    )
                    if self.stop_needed:
                        self.engineer("Pit required.")
        self.max_tyre_wear = current_max_wear

        if last_lap_time > 0:
            self.push({"total_time_ms": last_lap_time})
            self.printer.print_lap_time(previous_lap, last_lap_time, best)
            if self.tyre_data_emitted:
                self.printer.print_tyre(
                    self.session.tyre, self.session.tyre_age, rate_per_lap
                )
            if self.last_fuel is not None and self.session.fuel is not None:
                self.printer.print_fuel(self.session.fuel, self.last_fuel)
            self.last_fuel = self.session.fuel

        if current_lap:
            self.printer.print_lap(current_lap)

        self.tyre_data_emitted = False
        self.leader_distance.clear()
        self.leader_time.clear()

    def on_finish(self, lap, sectors, best):
        try:
            self.on_new_lap(None, lap, sectors, best)
        except Exception:
            pass
        print("🏁")

    def on_new_session(self, session):
        self.printer.print_session(session.slug)
        self.flush()

        self.queue.clear()

        self.motion_data = None
        self.rival_motion_data = None
        self.tyre_data_emitted = False

        self.last_live_data.clear()

        self.gap = 0.0  # meters
        self.leader_distance.clear()
        self.leader_time.clear()
        self.rival_index = 255
        self.distance = 0.0
        self.rival_distance = 0.0
        self.last_fuel = None

        if self.report:
            self.drivers.clear()
            self.human_count.clear()
            self.race_director = RaceDirector(self.drivers)

    # ---- PacketHandler ----

    def handle_SessionData(self, packet):
        self.session.refresh(packet)

        forecasts = [
            s
            for s in packet.weather_forecast_samples
            if s.time_offset > 0 and s.session_type == packet.session_type
        ]

        # Get the next session's forecast
        other_sessions_forecast = defaultdict(list)
        for f in (
            s
            for s in packet.weather_forecast_samples
            if s.time_offset > 0 and s.session_type != packet.session_type
        ):
            other_sessions_forecast[f.session_type].append(f)
        try:
            next_session_forecast = other_sessions_forecast[
                min(
                    _ for _ in other_sessions_forecast.keys() if _ > packet.session_type
                )
            ]
        except Exception:
            next_session_forecast = []

        if not forecasts:
            return

        _ = forecasts[0]
        next_forecast = (_weather(_)[1], _.time_offset, _.rain_percentage)
        # Only notify if the forecast changes significantly
        if (
            self._last_forecast is not None
            and next_forecast != self._last_forecast
            and (
                next_forecast[0] != self._last_forecast[0]
                or abs(next_forecast[-1] - self._last_forecast[-1]) > 10
            )
        ):
            self.engineer(
                r"{} in {} minutes with {}% chance of rain".format(*next_forecast)
            )
        self._last_forecast = next_forecast

        self.push_live(
            "weather_data",
            {
                "weather": _weather(packet),
                "forecasts": [
                    (sample.time_offset, *_weather(sample), sample.rain_percentage)
                    for _, sample in zip(
                        range(4),
                        forecasts,
                    )
                ],
                "nforecasts": [_weather(s)[0] for s in next_session_forecast],
            },
        )

    def handle_CarTelemetryData(self, packet: PacketCarTelemetryData):
        if self.motion_data is None:
            return

        try:
            data = packet.car_telemetry_data[_player_index(packet)].to_dict()
        except IndexError:
            return

        data.update(self.motion_data)
        if self.rival_motion_data:
            data.update(self.rival_motion_data)
            self.rival_motion_data = None
        self.motion_data = None

        data["distance"] = self.distance
        if self.session.type == SessionType.TT:
            data["gap"] = self.gap
            if self.rival_index != 255:
                rival = packet.car_telemetry_data[self.rival_index]
                data["rival_throttle"] = rival.throttle
                data["rival_brake"] = rival.brake
                data["rival_gear"] = rival.gear
                data["rival_speed"] = rival.speed
                data["rival_distance"] = self.rival_distance
                data["rival_steer"] = rival.steer

                self.push_live("trace", data)

        self.push_live("tyre_temp", data["tyres_inner_temperature"])

        for k, v in dict(data).items():
            if isinstance(v, list) and len(v) == len(TYRES):
                _flatten_tyre_values(data, k)

        self.push(data)

    def handle_CarStatusData(self, packet):
        try:
            data = packet.car_status_data[_player_index(packet)]
        except IndexError:
            return

        self.session.car_status_data(data)

        if not self.tyre_data_emitted:
            self.push(
                {
                    "tyre_compound": self.session.tyre,
                    "tyre_age": self.session.tyre_age,
                },
            )
            self.tyre_data_emitted = True

        self.push_live("fuel", data.fuel_remaining_laps)

        if self.last_fuel is None:
            self.last_fuel = data.fuel_remaining_laps

    def handle_CarDamageData(self, packet):
        try:
            data = packet.car_damage_data[_player_index(packet)].to_dict()
        except IndexError:
            return

        for k, v in dict(data).items():
            if isinstance(v, list) and len(v) == len(TYRES):
                _flatten_tyre_values(data, k)

        # Keep track of wing damage
        wing_status = data["front_left_wing_damage"], data["front_right_wing_damage"]
        if wing_status != self.wing_status and any(wing_status):
            self.engineer("Wing damage at {} {}".format(*wing_status))
        self.wing_status = wing_status

        self.push_live("car_status", data)
        self.push(data)

    def handle_FinalClassificationData(self, packet: PacketFinalClassificationData):
        self.flush()
        self.session.final_classification()

        if self.report:
            if self.session.is_qualifying():
                QualifyingReport(
                    self.drivers, packet.classification_data, self.human_count
                ).generate(f"{self.session.slug.replace('|', '_').replace(':', '')}-Q")
            elif self.session.is_race():
                RaceReport(
                    self.drivers, packet.classification_data, self.human_count
                ).generate(f"{self.session.slug.replace('|', '_').replace(':', '')}-R")
                if self.race_director is not None:
                    self.race_director.generate(
                        f"{self.session.slug.replace('|', '_').replace(':', '')}-D"
                    )
                    self.race_director = None
            else:
                print(f"No report generated for session type {self.session.type}")

    def handle_LapData(self, packet: PacketLapData):
        try:
            data = packet.lap_data[_player_index(packet)]
            self.rival_index = rival_index = packet.time_trial_rival_car_idx
            if rival_index != 255:
                rival = packet.lap_data[rival_index]
                gap = rival.lap_distance - data.lap_distance
                self.leader_timestamp.append(packet.header.session_time)
                if gap >= 0.0:
                    self.leader_distance.append(rival.lap_distance)
                    self.leader_time.append(rival.current_lap_time_in_ms)
                    i = closest(self.leader_distance, data.lap_distance)
                    self.gap = data.current_lap_time_in_ms - self.leader_time[i]
                else:
                    self.leader_distance.append(data.lap_distance)
                    self.leader_time.append(data.current_lap_time_in_ms)
                    i = closest(self.leader_distance, rival.lap_distance)
                    self.gap = self.leader_time[i] - rival.current_lap_time_in_ms
                self.gap /= 1000.0
                self.rival_distance = rival.lap_distance
            self.distance = data.lap_distance

            # Alert if negative SC delta
            if data.safety_car_delta < 0.0 and time() > self.sc_slow_down_timestamp:
                self.engineer(f"Delta {round(data.safety_car_delta, 1)} Slow down!")
                self.sc_slow_down_timestamp = time() + 5.0

        except IndexError:
            return

        self.session.lap_data(data)

    def handle_MotionData(self, packet):
        try:
            self.motion_data = packet.car_motion_data[_player_index(packet)].to_dict()
            if self.rival_index != 255:
                self.rival_motion_data = {
                    f"rival_{k}": v
                    for k, v in packet.car_motion_data[self.rival_index]
                    .to_dict()
                    .items()
                }
        except IndexError:
            return

    def handle_EventData(self, packet: PacketEventData):
        event = bytes(packet.event_string_code).decode()
        if event == "FLBK":
            # Flashback
            flashback_time = packet.event_details.flashback.flashback_session_time

            # Clean up leader data
            if self.leader_timestamp:
                i = closest(self.leader_timestamp, flashback_time)
                del self.leader_distance[i:]
                del self.leader_time[i:]
                del self.leader_timestamp[i:]

            # Remove events that are in the future w.r.t. the flashback time
            while self.queue:
                time, lap, _ = self.queue[-1]
                if time <= flashback_time:
                    self.session.lap = lap
                    break
                self.queue.pop()

            # We can flush the rest as we won't be flashing back beyond this
            # point in time.
            self.flush()
        elif event == "PENA":
            if self.race_director is not None:
                self.race_director.record_incident(packet.event_details.penalty)

    def handle_ParticipantsData(self, packet: PacketParticipantsData):
        if not self.report:
            return

        human_drivers = {
            i: player_name(p)
            for i, p in enumerate(packet.participants)
            if p.name and not p.ai_controlled
        }

        # Measure the "humanity" of the drivers
        self.human_count.update(human_drivers.values())

        self.drivers.update(human_drivers)

    def handle_generic(self, packet):
        self.last_packets[type(packet)] = packet

    def get_last(self, packet_type):
        return self.last_packets.get(packet_type)

    def engineer(self, message):
        def _():
            try:
                engine = pyttsx3.init()
                engine.say(message)
                engine.runAndWait()
            except:
                pass

        threading.Thread(target=_).start()



from datetime import datetime

from influxdb_client import InfluxDBClient
from influxdb_client import Point
from influxdb_client import WritePrecision
from influxdb_client.client.write_api import SYNCHRONOUS


class InfluxDBSinkError(Exception):
    pass


class InfluxDBSink:
    def __init__(self, org, token, bucket, url="http://localhost:8086"):
        self.client = InfluxDBClient(url=url, token=token, org=org, debug=False)
        try:
            self.client.ready()
            self._write_api = self.client.write_api(write_options=SYNCHRONOUS)
            self.bucket = bucket
        except:
            self.client = None

    def __enter__(self):
        if self.client is not None:
            self.client.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.client is not None:
            return self.client.__exit__(exc_type, exc_value, traceback)

    @property
    def connected(self):
        return self.client is not None

    def write(self, label, fields):
        if self.client is None:
            return

        p = Point(label)

        p._fields.update(fields)
        p.time(datetime.utcnow(), WritePrecision.MS)

        try:
            self._write_api.write(bucket=self.bucket, record=p)
        except Exception as e:
            # Best effort
            print(f"[ERROR] Cannot write to InfluxDB: {e}")



from datetime import timedelta


class SessionPrinter:
    def __init__(self):
        self.last_lap = 0
        self.last_sector = 0
        self.last_sectors = [None] * 4
        self.lap_ended = False

    def print_session(self, slug: str) -> None:
        print(f"Session {slug}")

    def print_lap(self, lap: int) -> None:
        if lap == self.last_lap:
            # Go back and clean the line
            print("\r" + " " * 60, end="\r", flush=True)
        else:
            print()
            self.last_sectors = [None] * 4

        print(f"Lap {lap:<6}", end="", flush=True)

        self.last_lap = lap
        self.last_sector = 0
        self.lap_ended = False

    def _reprint_sectors(self, n: int, lap: int) -> None:
        if not all(_ is None for _ in self.last_sectors):
            for sector in self.last_sectors[1:n]:
                if sector is not None:
                    self.print_sector(*sector)
                else:
                    print(" " * 11, end="", flush=True)
                    self.last_sector += 1

    def print_sector(
        self, n: int, lap: int, time: float, best_sector: bool, best_time: bool
    ) -> None:
        if lap != self.last_lap or n != self.last_sector + 1:
            # Refresh lap
            self.print_lap(lap)
            self.last_sectors[n + 1 :] = [None] * (3 - n)
            self.last_sectors[n] = (n, lap, time, best_sector, best_time)
            self._reprint_sectors(n, lap)
        else:
            self.last_sectors[n] = (n, lap, time, best_sector, best_time)

        if best_sector:
            bg = "105"
        elif best_time:
            bg = "102"
        else:
            bg = "103"

        print(
            f"\033[30;{bg}m" + f"{time/1000:02.3f}".center(10) + "\033[0m",
            end=" ",
            flush=True,
        )

        self.last_sector = n

    def print_lap_time(self, lap: int, time: float, best: float) -> None:
        if lap != self.last_lap or self.lap_ended:
            self.print_lap(lap)
            self._reprint_sectors(4, lap)

        time_repr = str(timedelta(milliseconds=time))[2:-3].center(13)
        if best:
            print(
                f"\033[95m" + f"{time_repr}".center(10) + "\033[0m", end="", flush=True
            )
        else:
            print(time_repr, end="", flush=True)

        self.lap_ended = True

    def print_tyre(self, tyre: str, tyre_age: int, wear_rate: int) -> None:
        t = tyre[0]
        color = {"S": "31", "M": "33", "H": "37", "I": "32", "W": 34}[t]
        print(
            f"\033[30;{color}m(\033[0m"
            + t
            + f"\033[30;{color}m)\033[0m  {tyre_age}  ({round(wear_rate)}% deg)",
            end="",
            flush=True,
        )

    def print_fuel(self, current_fuel: float, previous_fuel: float) -> None:
        current_color = "32" if current_fuel >= 0 else "31"
        delta_color = "32" if current_fuel - previous_fuel >= 0 else "31"

        print(
            f"  fuel \033[30;{current_color}m{current_fuel:.2f}\033[0m "
            f"(Δ: \033[30;{delta_color}m{current_fuel-previous_fuel:.2f}\033[0m)",
            end="",
            flush=True,
        )



from argparse import ArgumentParser
from threading import Thread

from f1.listener import PacketListener

from f1_telemetry import live
from f1_telemetry.collector import TelemetryCollector
from f1_telemetry.server import serve
from f1_telemetry.storage import InfluxDBSink
from f1_telemetry.storage import InfluxDBSinkError


DEFAULT_BUCKET = "f1-telemetry"


def main():
    argp = ArgumentParser(prog="f1-tel")

    argp.add_argument(
        "org",
        help="InfluxDB Org",
        type=str,
    )
    argp.add_argument(
        "token",
        help="InfluxDB Token",
        type=str,
    )
    argp.add_argument(
        "-b",
        "--bucket",
        help="InfluxDB Bucket",
        type=str,
        default=DEFAULT_BUCKET,
    )
    argp.add_argument(
        "-a",
        "--address",
        help="F1 game telemetry address",
        type=str,
        default="localhost",
    )
    argp.add_argument(
        "-p",
        "--port",
        help="F1 game telemetry port",
        type=int,
        default=20777,
    )
    argp.add_argument(
        "-H",
        "--host",
        help="Server host address",
        type=str,
        default="localhost",
    )
    argp.add_argument(
        "-r",
        "--report",
        help="Generate reports at the end of sessions. Useful for session coordinators",
        action="store_true",
    )

    args = argp.parse_args()

    collector = None

    try:
        with InfluxDBSink(org=args.org, token=args.token, bucket=args.bucket) as sink:
            if not sink.connected:
                print(
                    "WARNING: InfluxDB not available. Telemetry data will not be stored."
                )
            else:
                print("Connected to InfluxDB")

            listener = PacketListener(host=args.address, port=args.port)
            collector = TelemetryCollector(listener, sink, args.report)

            server_thread = Thread(target=serve, args=(args.org, args.token, args.host))
            server_thread.daemon = True
            server_thread.start()

            print("Listening for telemetry data ...")
            collector_thread = Thread(target=collector.collect)
            collector_thread.daemon = True
            collector_thread.start()

            print("Starting live data websocket server")
            # FIXME: Mixing asyncio and threads is yuck!
            live.serve(host=args.host)

    except InfluxDBSinkError as e:
        print("Error:", e)

    except KeyboardInterrupt:
        if collector is not None:
            collector.flush()
        print("\nBOX BOX.")


if __name__ == "__main__":
    main()



import csv
import typing as t
from collections import Counter
from datetime import timedelta

from f1.packets import FinalClassificationData
from f1.packets import Penalty


class HumanCounter(Counter):
    def __init__(self, *args, **kwargs):
        self.total = -1  # Update is called on instance creation
        super().__init__(*args, **kwargs)

    def update(self, *args, **kwargs):
        self.total += 1
        super().update(*args, **kwargs)

    def clear(self):
        self.total = 0
        super().clear()

    def get_percents(self):
        return {
            name: 100.0 * count / self.total if self.total > 0 else 0
            for name, count in self.items()
        }


def result_status(data: FinalClassificationData) -> str:
    return {
        0: "INV",
        1: "INA",
        2: "ACT",
        3: "FIN",
        4: "DNF",
        5: "DQ",
        6: "NC",
        7: "DNF",  # "RET",
    }[data.result_status]


def fmtt(millis: int) -> str:
    result = str(timedelta(milliseconds=millis))[:-3]
    if result.startswith("0:"):
        return result[2:]
    return result


class Report:
    def __init__(
        self,
        drivers: t.Dict[int, str],
        data: t.List[FinalClassificationData],
        human_count: HumanCounter,
    ) -> None:
        self.drivers = drivers
        self.data = data
        self.human_pct = human_count.get_percents()

    def rows(self) -> t.Iterator[t.Tuple]:
        pass

    def generate(self, report_name: str) -> None:
        with open(
            f"{report_name}.csv", "w", newline="", encoding="utf-8"
        ) as final_file:
            writer = csv.writer(final_file)
            for row in self.rows():
                writer.writerow(row)


class QualifyingReport(Report):
    def rows(self):
        final_data = []
        for i, name in self.drivers.items():
            data = self.data[i]
            final_data.append(
                (
                    data.position,
                    name,
                    data.best_lap_time_in_ms,
                    fmtt(data.best_lap_time_in_ms),
                )
            )

        return sorted(final_data, key=lambda x: x[0])


class RaceReport(Report):
    def rows(self):
        final_data = []
        for i, name in self.drivers.items():
            data = self.data[i]

            total_time = int((data.total_race_time + data.penalties_time) * 1000)

            final_data.append(
                (
                    data.position,
                    name,
                    data.best_lap_time_in_ms,
                    fmtt(data.best_lap_time_in_ms),
                    total_time,
                    fmtt(total_time),
                    data.num_laps,
                    result_status(data),
                    data.penalties_time,
                    round(self.human_pct.get(name, 0.0)),
                )
            )

        return sorted(final_data, key=lambda _: _[0])


class RaceDirector(Report):
    INCIDENT_TYPES = {
        0: "Drive through",
        1: "Stop Go",
        2: "Grid penalty",
        3: "Penalty reminder",
        4: "Time penalty",
        5: "Warning",
        6: "Disqualified",
        7: "Removed from formation lap",
        8: "Parked too long timer",
        9: "Tyre regulations",
        10: "This lap invalidated",
        11: "This and next lap invalidated",
        12: "This lap invalidated without reason",
        13: "This and next lap invalidated without reason",
        14: "This and previous lap invalidated",
        15: "This and previous lap invalidated without reason",
        16: "Retired",
        17: "Black flag timer",
    }

    INCIDENT_DETAILS = {
        0: "Blocking by slow driving",
        1: "Blocking by wrong way driving",
        2: "Reversing off the start line",
        3: "Big Collision",
        4: "Small Collision",
        5: "Collision failed to hand back position single",
        6: "Collision failed to hand back position multiple",
        7: "Corner cutting gained time",
        8: "Corner cutting overtake single",
        9: "Corner cutting overtake multiple",
        10: "Crossed pit exit lane",
        11: "Ignoring blue flags",
        12: "Ignoring yellow flags",
        13: "Ignoring drive through",
        14: "Too many drive throughs",
        15: "Drive through reminder serve within n laps",
        16: "Drive through reminder serve this lap",
        17: "Pit lane speeding",
        18: "Parked for too long",
        19: "Ignoring tyre regulations",
        20: "Too many penalties",
        21: "Multiple warnings",
        22: "Approaching disqualification",
        23: "Tyre regulations select single",
        24: "Tyre regulations select multiple",
        25: "Lap invalidated corner cutting",
        26: "Lap invalidated running wide",
        27: "Corner cutting ran wide gained time minor",
        28: "Corner cutting ran wide gained time significant",
        29: "Corner cutting ran wide gained time extreme",
        30: "Lap invalidated wall riding",
        31: "Lap invalidated flashback used",
        32: "Lap invalidated reset to track",
        33: "Blocking the pitlane",
        34: "Jump start",
        35: "Safety car to car collision",
        36: "Safety car illegal overtake",
        37: "Safety car exceeding allowed pace",
        38: "Virtual safety car exceeding allowed pace",
        39: "Formation lap below allowed speed",
        40: "Formation lap parking",
        41: "Retired mechanical failure",
        42: "Retired terminally damaged",
        43: "Safety car falling too far back",
        44: "Black flag timer",
        45: "Unserved stop go penalty",
        46: "Unserved drive through penalty",
        47: "Engine component change",
        48: "Gearbox change",
        49: "Parc Fermé change",
        50: "League grid penalty",
        51: "Retry penalty",
        52: "Illegal time gain",
        53: "Mandatory pitstop",
        54: "Attribute assigned",
    }

    def __init__(self, drivers):
        super().__init__(drivers, [], HumanCounter())

    def record_incident(self, penalty: Penalty) -> None:
        if penalty.vehicle_idx not in self.drivers:
            return

        try:
            self.data.append(
                (
                    self.drivers[penalty.vehicle_idx],
                    self.INCIDENT_TYPES[penalty.penalty_type],
                    self.INCIDENT_DETAILS[penalty.infringement_type],
                    penalty.lap_num,
                    self.drivers.get(penalty.other_vehicle_idx, ""),
                )
            )
        except KeyError:
            print(
                "ERROR: Unknown penalty type:",
                penalty.penalty_type,
                penalty.infringement_type,
            )

    def rows(self):
        return self.data
